#!/usr/bin/env python
import traceback
import sys,glob
import os.path

import numpy
from datetime import date
import netCDF4
import time
from numba import *
from utils import *
from anomaly import *
#import matplotlib.pylab as plt
import scipy.stats

plt.rcParams['lines.linewidth'] = 3

def read_temperature(stats,fn,tidx,varlist=['temperatures','fg_dep','an_dep']):
    
    if 'mdatum' not in stats.keys():
        try:
            f=netCDF4.Dataset(fn,'r')
            f.set_auto_mask(False)
            s=getattr(f.variables['datum'],'units').split()[2].split('-')[0]
            shift=tidx[(int(s)-1900)*12]
            stats['mdatum']=f.variables['datum'][0]+shift
            stats['mhours']=f.variables['hours'][:]
	    stats['lat']=f.variables['lat'][0]
	    stats['lon']=f.variables['lon'][0]
            for var in varlist:
		try:
		    miss_val=getattr(f.variables['temperatures'],'missing_value')
		    vals=f.variables[var][:]
		    vals[vals==miss_val]=numpy.nan
		except KeyError:
		    vals=f.variables['temperatures'][:]+numpy.nan
		if var not in ['fg_dep','an_dep']:
		    stats['m'+var]=vals
		else:
		    stats['m'+var]=-vals
            if 'merged' not in fn:
## departures in merged files are defined as bg-obs, not obs-bg, thus the sign change
                b=f.variables['bias'][:]
            else:
                b=stats['mtemperatures']-stats['mtemperatures']
             
            b[numpy.isnan(stats['mtemperatures'])]=numpy.nan   
            stats['mbias']=-b
            stats['mtemperatures']-=b
            stats['m'+fgdepname]=b-stats['m'+fgdepname]
            stats['man_dep']=b-stats['man_dep']
                
            f.close()
        except:
            p=False
            pass
    else:

#   merge
        ntimes=list()
        try:
            f=netCDF4.Dataset(fn,'r')
            f.set_auto_mask(False)
            if 'datum' in f.variables.keys():
                s=getattr(f.variables['datum'],'units').split()[2].split('-')[0]
                shift=tidx[(int(s)-1900)*12]
                times=f.variables['datum'][0]+shift
                ntimes=numpy.zeros(indexmax,dtype=numpy.int)
#                ntimes[times]=1
                ntimes[stats['mdatum']]=1
                nhours=numpy.zeros([2,indexmax],dtype=numpy.int)-999
                try: 
                    nhours[:,times]=f.variables['hours'][:]
                except KeyError:
                    pass
                nhours[:,stats['mdatum']]=stats['mhours'][:]
                t=time.time()
                mdathilf=numpy.asarray(numpy.where(ntimes>0)[0],dtype=numpy.int)
                
                if 'mtemperatures' not in stats.keys():
		    return
		for var in varlist:
                    mergevar(stats,fn,f.variables,times,mdathilf,var)
                    
                #print 'merge',time.time()-t
                stats['mdatum']=mdathilf
                stats['mhours']=nhours[:,mdathilf]
            f.close()
        except:
            p=False
            pass
        
def mergevar(stats,fn,variables,times,mdathilf,var):
            temperatures=numpy.empty([2,stats['mtemperatures'].shape[1],indexmax],dtype=numpy.float32)
            miss_val=getattr(variables[var],'missing_value')
            vals=variables[var][:]
            
            if 'corr' in fn:
		timeoffset=daysbetween(19000101,variables["datum"].units)        
                index=variables['datum'][0,:]-1+timeoffset
		stats['index']=index
#                index=numpy.arange(len(index),dtype=int)   
                try:
                    nc_miss_val=variables[var].getncattr('_FillValue')
                except:
                    pass
                try:
                    nc_miss_val=variables[var].getncattr('missing_value')
                except:
                    pass
            
		
		temperatures[:]=0.
		hilf=variables[var][:]
		hilf[hilf==nc_miss_val]=numpy.nan
		expand2(hilf,index,numpy.arange(variables[var].shape[1],dtype='int'),temperatures)
#                copystride(temperatures,variables[var][:],index,0,0,
#                           numpy.arange(variables[var].shape[1],dtype='int'),nc_miss_val)
		vals=temperatures.copy()

	    else:
		temperatures[:]=numpy.nan
		vals[vals==miss_val]=numpy.nan
		if 'merged' not in fn:
		    try:
			b=variables['bias'][:]
		    except KeyError:
			b=vals-vals
		else:
		    b=vals-vals
    # departures in merged files are defined as bg-obs, not obs-bg, thus the sign change
		if var=='temperatures':
		    vals-=b
		elif var=='fg_dep':
		    vals=b-vals
		elif var=='an_dep':
		    vals=b-vals
    
		temperatures[:,:,times]=vals
	    if 'rio' in fn:
		stats['mrich'+var]=temperatures[:,:,mdathilf]
	    else:
		stats['m'+var]=temperatures[:,:,mdathilf]
		

@njit
def calc_bin_and_chunk(full,allindices,i,bins,weights):
    
    chunko=full[:,:,allindices[i-1]:allindices[i]]
    chunk=numpy.empty(chunko.shape)
    chunk[:]=chunko[:]
    goods=numpy.zeros((12,chunko.shape[0],chunko.shape[1]),dtype=numpy.int32)
    for j in range(12):
	for l in range(bins.shape[2]):
	    for m in range(bins.shape[3]):
		bins[i-1,j,l,m]=0.
		kk=0
		for ii in range(allindices[i-1],allindices[i]):
                    if ii<date_listm.shape[0]:
                        if date_listm[ii]==j+1:
                            if chunk[l,m,ii-allindices[i-1]]==chunk[l,m,ii-allindices[i-1]]:
                                bins[i-1,j,l,m]+=chunk[l,m,ii-allindices[i-1]]
                                kk+=1
		goods[j,l,m]=kk
		if kk>5:
		    bins[i-1,j,l,m]/=kk
		    for ii in range(allindices[i-1],allindices[i]):
                        if ii<date_listm.shape[0]:
			    if date_listm[ii]==j+1:
			        if chunk[l,m,ii-allindices[i-1]]==chunk[l,m,ii-allindices[i-1]]:
				    chunk[l,m,ii-allindices[i-1]]-=weights[m]*bins[i-1,j,l,m]
		else:
		    bins[i-1,j,l,m]=numpy.nan
		    for ii in range(allindices[i-1],allindices[i]):
                        if ii<date_listm.shape[0]:
			    if date_listm[ii]==j+1:
			        chunko[l,m,ii-allindices[i-1]]-=numpy.nan
			
    return bins,goods,chunk,chunko


if __name__ == "__main__":


    if len(sys.argv)<2:
	exp='exp01'
    else:
	exp=sys.argv[1]
    path='/home/srvx7/leo/fastscratch/rise/1.0/'
    
    fgdepname='eijra_fgdep'  # This is basically the ERA-Interim background departure
    
    f=netCDF4.Dataset(path+exp+'/ERA5bc_RAOBCORE_v1.5_template.nc','r')
#    f=netCDF4.Dataset('/home/srvx7/raobcore/v1.5.1/export/ERA5/ERA5bc_RAOBCORE_v1.5_098646.nc','r')
    f.set_auto_mask(False)
    vals=f.variables['temperatures'][1,3,:]
    offsets=f.variables['datum'][0,:]
    print vals[:4],offsets[0]
    print datetime.date(1900,1,1)+datetime.timedelta(days=int(offsets[0]))
    mask=vals!=-999.
    
    plpath=os.path.expandvars('plots/')
    try:
	os.mkdir(plpath)
    except:
	pass
    monthdays=[31,28,31,30,31,30,31,31,30,31,30,31]  
    start=1900
    stop=2018
    plevs=numpy.asarray([10,20,30,50,70,100,150,200,250,300,400,500,700,850,925,1000])
    nlev=plevs.shape[0]
    
    stats=dict()
    lats=list()
    lons=list()
    indexmax=45000
    tidx=calcdays(19000101,(2018-1900)*12)-1
    l=0
    t=time.time()
    forig=glob.glob(path+exp+'/[0-9]*/feedbackmerged05????.nc')
    ic=0
    mlist=[]
    for line in forig:
	st=line.split('/')[-1][-9:-3]
	mlist.append(st)
        ic+=1
	
    base = datetime.date(1900,1,1)
    
    il=0
    for line in forig:
	st=line.split('/')[-1][-9:-3]
	fns=[path+exp+'/'+st+'/feedbackmerged'+st+'.nc',
	    path+exp+'/'+st+'/'+'feedbackglobbincorrsave'+st+'.nc'
	    ]
	varlists=[['temperatures',fgdepname,'an_dep'],['rasocorr']]
	i=0
	stats[st]=dict()
	for fn in fns:
	    read_temperature(stats[st],fn,tidx,varlist=varlists[i])
	    i+=1
    
        if 'lat' not in stats[st].keys() or 'mdatum' not in stats[st].keys():
	    print st,'no data found,cycling ..'
	    continue
	s=stats[st]
	
	t2=time.time()
	elev=calc_elevangles(0.,numpy.asarray([s['lat']]),numpy.asarray([s['lon']]))
	for y in range(120):
	    idx=datetime.date(1900+y,1,1)-base
	    idx=idx.days
	    elev[0,:,idx:idx+366]=elev[0,:,:366]
	print 't2:',time.time()-t2
    
	bdatum=[]
	try:
	    if il==0:
		full=numpy.empty([2,s['mtemperatures'].shape[1],indexmax],dtype=numpy.float32)
		fullc=numpy.empty([2,s['mtemperatures'].shape[1],indexmax],dtype=numpy.float32)
		il=1
	    smask=numpy.abs(s['mtemperatures'])<400.
	except:
	    print st,' could not be read'
	    continue
	full.fill(numpy.nan)
	
	if numpy.sum(smask)>0:
	    print st, numpy.sum(smask), ' values found'
	smask=numpy.abs(s['m'+fgdepname])>20.
	s['m'+fgdepname][smask]=numpy.nan
	s['mtemperatures'][smask]=numpy.nan
	s['mfg']=s['mtemperatures']+s['m'+fgdepname]
	
	full[:,:,s['mdatum']]=s['m'+fgdepname]
	fullc[:]=full[:]
	allindices=[s['mdatum'][0]]
	try:
	    for idx in s['index']:
		if idx>s['mdatum'][0] and idx!=20819:
		    allindices.append(idx)
	    if s['index'][-1]<s['mdatum'][-1]:
		allindices+=[s['mdatum'][-1]]
	    else:
		allindices[-1]=s['mdatum'][-1]
	except KeyError:
	    print st,'no break found'
	    allindices.append(s['mdatum'][-1])
	
	date_list = [base + datetime.timedelta(days=x) for x in range(0, allindices[-1]+1)]
	date_listm = numpy.asarray([date_list[i].month for i in range(0, allindices[-1]+1)])
	date_listy = numpy.asarray([date_list[i].year for i in range(0, allindices[-1]+1)])
    
	bins=numpy.empty((len(allindices),12,s['m'+fgdepname].shape[0],s['m'+fgdepname].shape[1]))
	svar=numpy.empty((len(allindices),bins.shape[2],bins.shape[3]),dtype='bool')
	sratio=numpy.empty((len(allindices),bins.shape[2],bins.shape[3]),dtype='float32')
	scorr=numpy.empty((len(allindices),bins.shape[2],bins.shape[3]),dtype='float32')
	svar[:]=False
	weights=numpy.asarray((0.9,0.9,0.9,0.9,0.9,0.9,0.8,0.8,0.7,0.6,0.5,0.3,0.01,0.01,0.0,0.0))
	allindices=numpy.array(allindices)
	for i in range(1,len(allindices)):
    
	    t=time.time()
	    bins,goods,chunk,chunko=calc_bin_and_chunk(full,allindices,i,bins,weights)
	    print time.time()-t
	    
	    for k in range(chunk.shape[0]):
		for l in range(chunk.shape[1]):
		    if sum(goods[:,k,l]>5)>6:
			c=chunk[k,l,:]-chunko[k,l,:]
			mask=~numpy.isnan(c)
			if sum(mask)>0:
			    selev=elev[0,k,allindices[i-1]:allindices[i]]
			    if mask.shape[0]>366:
				sratio[i-1,k,l]=numpy.std(c[mask][:366])/numpy.std(chunk[k,l,mask][:366])
				scorr[i-1,k,l]=scipy.stats.spearmanr(c[mask][:366],selev[mask][:366]).correlation
			    else:
				sratio[i-1,k,l]=numpy.std(c[mask])/numpy.std(chunk[k,l,mask])
				scorr[i-1,k,l]=scipy.stats.spearmanr(c[mask],selev[mask]).correlation
			    svar[i-1,k,l]=sratio[i-1,k,l]>0.1
			else:
			    scorr[i-1,k,l]=numpy.nan
			    sratio[i-1,k,l]=numpy.nan
		    else:
			scorr[i-1,k,l]=numpy.nan
			sratio[i-1,k,l]=numpy.nan
		if sum(svar[i-1,k,:])>5:
		    svar[i-1,k,:]=True
		else:
		    svar[i-1,k,:]=False
		    
			    
	    for k in range(chunk.shape[0]):
		for l in range(chunk.shape[1]-4):  # no RISE adjustment below 500 hPa
		    if sum(goods[:,k,l]>5)>6:
			c=chunk[k,l,:]-chunko[k,l,:]
			if svar[i-1,k,l]:
			    fullc[k,l,allindices[i-1]:allindices[i]]=chunk[k,l,:]-numpy.nanmean(chunk[k,l,:]-chunko[k,l,:])
			if numpy.nanmax(numpy.abs(chunk[k,l,:]-chunko[k,l,:]))>5:
			    print k,l,allindices[i-1]/365.25+1900,numpy.nanmax(numpy.abs(chunko[k,l,:]-chunk[k,l,:])),numpy.nanmean(chunk[k,l,:]-chunko[k,l,:])

	    print 'f:',time.time()-t
	s['elev']=elev[0,:,s['mdatum']]
	print 't2:',time.time()-t2
	#continue
    
	
	if st in mlist:
	    pass
	else:
	    print st,s['mtemperatures'].shape
	    pass
	    
	if 'mbias' not in s.keys():
	    print st+' no RASE bias'
	    continue
	if 'mtemperatures' not in s.keys():
	    print st+' not plotted'
	    print s['mbias'].shape
	    continue
	#if 'mRAOBCORE' not in s.keys():
	    #s['mRAOBCORE']=s['mtemperatures']-s['mtemperatures']
	    #print st+' mRAOBCORE not found, zero correction supplied'
	if 'mrasocorr' not in s.keys() or 'mrasocorr' not in s.keys():
	    s['mrasocorr']=s['mtemperatures']-s['mtemperatures']
	    print st+' mrasocorr not found, zero correction supplied, cycle'
	    continue

	t=time.time()
	try:
	    s['newbias']=s['mrasocorr']
#	    s['newrichbias']=s['mrichrasocorr']
	except:
	    print 'no bias corrections in file, cycling'
	    continue
    
	try:
	    x1979=daysbetween(19000101,'days since 1979-01-01 00:00:00')
	    idx=numpy.where(s['mdatum']>x1979)
	    if len(idx[0])>0:
#		s['newrichbias'][:,:,idx[0]]=s['newrichbias'][:,:,idx[0]]+fullc[:,:,s['mdatum'][idx]]-full[:,:,s['mdatum'][idx]]
		s['newbias'][:,:,idx[0]]=s['newbias'][:,:,idx[0]]+fullc[:,:,s['mdatum'][idx]]-full[:,:,s['mdatum'][idx]]
	    pass
	except:
#	    s['newrichbias']=fullc[:,:,s['mdatum']]-full[:,:,s['mdatum']]
	    s['newbias']=fullc[:,:,s['mdatum']]-full[:,:,s['mdatum']]
	    print st,'no adjustments'
	    pass
	
	fig=True
	pindex=[5]
    #    pindex=[0,1,2,3,4,5,6,7,8,9,10,11,12,13]
	nm=numpy.nanmax(numpy.abs(s['newbias']))
	if nm>7:
	    print st,'nm:',nm
	    pindex=[0,1,2,3,4,5,6,7,8,9,10,11,12,13]
		
    
	if 'mtemperatures' in s.keys():
	    flag=False
	    fn=path+exp+'/'+st+'/feedbackmerged'+st+'.nc'
	    f = netCDF4.Dataset(fn,"r")
	    fno=path+exp+'/'+st+'/ERA5bc_RAOBCORE_v1.6_'+st+'.nc'
	    fo = netCDF4.Dataset(fno,"w", format='NETCDF4_CLASSIC')
	      
	    for i in f.ncattrs():
		if i=='history':
		    setattr(fo,i,datetime.date.today().strftime("%Y/%m/%d"))
		elif i=='source':
		    setattr(fo,i,'RAOBCORE/RICH v1.6 + solar elevation dependency (from 197901 onward)' )
		elif i=='title':
		    setattr(fo,i,'Station daily temperature series with ERA-Interim background departure statistics and RISE bias estimates' )
		else:
		    setattr(fo,i,getattr(f,i))
	    for i in f.dimensions.keys():
		if i=='time':
		    fo.createDimension(i,s["mdatum"].shape[0])
		else:
		    try:
			fo.createDimension(i,len(f.dimensions[i]))
		    except:
			flag=True
			continue
	    nalias=8    
	    fo.createDimension('nalias',8)
	    fo.createDimension('nchar',8)
	    if flag:
		continue
	    #nogos=['flags',u's_type', u'eijra_fgdep', u'jra55_fgdep', u'jra55_andep', u'e20c_andep', u'n20c_andep', u'ce20c_andep', u'erapresat_andep']
	    tobecopied=['datum','hours','lat','lon','alt','press','temperatures','an_dep','fg_dep','source','mergedstats']
	    for i in f.variables.keys():
		var=f.variables[i]
		if i=='datum':
		    fo.createVariable(i,var.dtype,var.dimensions)
		    fo.variables[i][:]=s["mdatum"][:]
		elif i=='hours':
		    fo.createVariable(i,var.dtype,var.dimensions)
		    fo.variables[i][:]=s["mhours"][:]
		elif i=='temperatures':
		    fo.createVariable(i,var.dtype,var.dimensions)
		    s["mtemperatures"][numpy.isnan(s["mtemperatures"])]=-999.
		    fo.variables[i][:]=s["mtemperatures"][:]
		elif i=='fg_dep':
		    fo.createVariable(i,var.dtype,var.dimensions)
		    s["m"+fgdepname][numpy.isnan(s["m"+fgdepname])]=-999.
		    fo.variables[i][:]=s["m"+fgdepname][:]
		elif i=='an_dep':
		    s["newbias"][numpy.isnan(s["newbias"])]=999.
#		    s["newrichbias"][numpy.isnan(s["newrichbias"])]=999.
		    try:
			fo.createVariable('bias',var.dtype,var.dimensions)
		    except:
			pass
		    fo.variables['bias'][:]=-s["newbias"][:]
#		    fo.createVariable('richbias',var.dtype,var.dimensions)
#		    fo.variables['richbias'][:]=-s["newrichbias"][:]
		elif i in ('lon','lat','alt','press'):
		    fo.createVariable(i,var.dtype,var.dimensions)
		    fo.variables[i][:]=var[:]
		elif i in ('source'):
#		    str_out = netCDF4.stringtochar(numpy.array(['test'], 'S4'))
		    fo.createVariable(i,'S1',('time','nchar'))
		    x=numpy.empty((var.shape[0]),dtype='S8')
		    x.fill('        ')
		    try:
			svar=var[:]
		    except:
			print 'could not read source, supplying BUFRDATA'
			svar=x
		    str_out = netCDF4.stringtochar(numpy.asarray(svar,'S8'))
		    fo.variables[i][:]=str_out
		elif i in ('mergedstats'):
		    fo.createVariable(i,'S1',('time','nalias','nchar'))
		    tt=time.time()
		    x=numpy.empty((var.shape[0],nalias),dtype='S8')
		    x.fill('        ')
		    try:
			svar=var[:]
			for k in range(svar.shape[0]):
				l=svar[k].split(',')
				for m in range(len(l)):
				    x[k,m]=l[m]+' '*(8-len(l[m]))
		    except:
			print 'could not read mergedstats, filling with WMO number'
			x.fill(st[1:]+'   ')

		    str_out = netCDF4.stringtochar(x)
		    fo.variables[i][:]=str_out
		    print 'mergevar:',time.time()-tt
		else:
		    if i in tobecopied:
			print i,'some unknown variable'
			fo.createVariable(i,var.dtype,var.dimensions)
	  
		for j in var.ncattrs():
		    if j!='_FillValue' and j!='scale_factor' and j!='add_offset':
			if i in tobecopied:
			    if i=='an_dep':
				setattr(fo.variables['bias'],j,getattr(var,j))
#				setattr(fo.variables['richbias'],j,getattr(var,j))
			    elif i=='datum' and j=='units':
				setattr(fo.variables[i],j,'days since 1900-01-01 00:00:00')
			    else:
				setattr(fo.variables[i],j,getattr(var,j))
				
	    setattr(fo.variables['fg_dep'],'infos','obs-CERA20C ensemble average up to 197812; obs-ERA_Interim 197901 onwards')
	    setattr(fo.variables['bias'],'long_name','RAOBCORE v1.6 RISE bias estimate')
#	    setattr(fo.variables['richbias'],'long_name','RICH v1.6 RICH bias estimate')
	    setattr(fo.variables['source'],'info','Preferred source used for merged temperature record')
	    setattr(fo.variables['source'],'valid_entries','BUFRDATA: (ECMWF holdings), NCARUA(20,21,24): sources from NCAR, CHUAN2.1: ERA-CLIM(2) digitized data')
	    setattr(fo.variables['mergedstats'],'info','ODB statIDs that matched during merge - bias adjustments should be applied to those')
	  
       
	    fo.close()
	    f.close()
    
    print time.time()-t
